/************************************************************
 *  Martin Toepfer     ***     PROJECT : Lingvisto          *
 *  Wissmannstr. 5     ***                                  *
 *  D - 22041 Hamburg  ***     File    : edit.c             *
 *  Germany            ***     Date    : 04.04.1997         *
 ************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "datastructure.h"
#include "edit.h"
#include "error.h"

/************************************************************
 *        global variables (in datastructure.c)             *
 ************************************************************/  
extern struct lvo_file head;
extern NODE current;

/**********************************************************************
 *                           edit functions                           *
 **********************************************************************/
/************************************************************
 * insert_pair : insert NEW pair of words into list         *
 *                                                          *
 * native  : native word to insert                          *
 * foreign : foreign word to insert                         *
 ************************************************************/
void insert_pair(char *native, char *foreign) {
  NODE n = n_alloc();
  if (n != NULL)  {
    set_state(CHANGED);                  /* unsaved changes */
    set_words(native, foreign, n);
    
    /*
     * insert node into wordlist
     */

    if (current == NULL) {                /* no node exists */
      n->next = NULL;
      n->prev = NULL;
      set_wordlist(n);
    }
    else {	                /* at least one node exists */
      n->next = current->next;
      n->prev = current;
      current->next = n;
      if (n->next != NULL)         /* node is not last node */
	n->next->prev = n;
    }
    current = n;
  }
  else
    lvo_error(NO_MEMORY);
}

/************************************************************
 * delete_pair : delete pair of words from list             *
 ************************************************************/
void delete_pair(void) {
  if (current != NULL)  {
    NODE help = current;
    set_state(CHANGED);                  /* unsaved changes */
    
    /*
     * mid node
     */
    
    if ((current->next != NULL) && (current->prev != NULL)) {
      current->next->prev = current->prev;
      current->prev->next = current->next;
      current = current->next;
    }
    
    /*
     * last node
     */
    
    else if ((current->next == NULL) && (current->prev != NULL)) {
      current->prev->next = NULL;
      current = current->prev;
    }
    
    /*
     * first node
     */
    
    else if ((current->next != NULL) && (current->prev == NULL)) {
      current->next->prev = NULL;
      set_wordlist(current->next);
      current = current->next;      
    }
    
    /*
     * sole node
     */
    
    else {
      current = NULL;
      set_wordlist(current);
    }
    n_free(help);
  }
  else
    lvo_error(DELETE_EMPTY_LIST);
}

/************************************************************
 * change_pair : change current pair to NEW pair of words   *
 *                                                          *
 * native  : native word that replaces old word             *
 * foreign : foreign word that replaces old word            *
 ************************************************************/
void change_pair(char *native, char *foreign) {
  set_state(CHANGED);                    /* unsaved changes */
  free(current->pair.native);
  free(current->pair.foreign);
  set_words(native, foreign, current);
}

/************************************************************
 * find_pair : try to find pair of words                    *
 *                                                          *
 * s      : search-string                                   *
 * mode   : search mode (FOREIGN or NATIVE)                 *
 ************************************************************/
void find_pair(char *s, unsigned char mode) {
  NODE tmp;
  
  /*
   * search substring in 'mode' language from current
   * position in wordlist to end of wordlist
   */
  
  for(tmp = current; tmp != NULL; tmp = tmp->next) {
    if (((mode == NATIVE) && (strstr(tmp->pair.native, s) != NULL)) ||
	((mode == FOREIGN) && (strstr(tmp->pair.foreign, s) != NULL))) {
      current = tmp;
      return;                                      /* match */
    }
  }
  fprintf(stderr,"\a");                         /* no match */
}

/************************************************************
 * forward_pair : set current to next pair of words         *
 ************************************************************/
void forward_pair(void)  {
  if (current != NULL) {
    if (current->next != NULL)
      current = current->next;
    else
      fprintf(stderr, "\a");             /* end of wordlist */
  }
}

/************************************************************
 * reverse_pair : set current to previous pair of words     *
 ************************************************************/
void reverse_pair(void) {
  if (current != NULL) {
    if (current->prev != NULL)
      current = current->prev;
    else
      fprintf(stderr, "\a");           /* begin of wordlist */
  }
}

/************************************************************
 * delete_all : delete all pairs                            *
 ************************************************************/
void delete_all(void) {
  while (current != NULL)
    delete_pair();
  set_state(UNCHANGED);
}

/**********************************************************************
 *                      Xaw functions                                 *
 **********************************************************************/
#include <setjmp.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include "xawfunction.h"

/************************************************************
 *             global variables (in xaw.c)                  *
 ************************************************************/
extern jmp_buf env1;
extern String trans1, trans3, transFWD, transREV;
extern Widget form, native, foreign, test;

/************************************************************
 * InsertPair : inserts pair of words                       *
 ************************************************************/
void InsertPair(void) {
  char *n;
  char *f;
  SetAllSensitive((Boolean) False);
  while (setjmp(env1) == 0) {
    XtVaSetValues(native, XtNstring, "", NULL);
    XtVaSetValues(foreign, XtNstring, "", NULL);
    
    /*
     * read native language word
     */
    
    DisplayState("INSERT : Please insert new NATIVE word !");
    XtOverrideTranslations(native,
			   XtParseTranslationTable(trans3));
    n = GetString(native);
    XtOverrideTranslations(native,
			   XtParseTranslationTable(trans1));
    XtVaSetValues(native, XtNstring, n, NULL);
    
    /*
     * read foreign language word
     */
    
    DisplayState("INSERT : Please insert new FOREIGN word !");
    XtOverrideTranslations(foreign,
			   XtParseTranslationTable(trans3));
    f = GetString(foreign);
    XtOverrideTranslations(foreign,
			   XtParseTranslationTable(trans1));
    
    /*
     * insert pair
     */
    
    insert_pair(n, f);
  }
}

/************************************************************
 * ChangePair : change pair of words                        *
 ************************************************************/
/**************************************************
 *          callback of 'apply' button            *
 **************************************************/
/* ARGSUSED */
static void ApplyChanges(Widget w, XtPointer x, XtPointer y) {
  String s1, s2;
  char *n;
  char *f;

  /*
   * read changed foreign and native language word
   */
  
  XtVaGetValues(native, XtNstring, &s1, NULL);
  XtVaGetValues(foreign, XtNstring, &s2, NULL);
  n = (char *) malloc(strlen(s1) + 1);
  f = (char *) malloc(strlen(s2) + 1);
  if ((n != NULL) && (f != NULL)) {
    strcpy(n, s1);
    strcpy(f, s2);
    change_pair(n, f);
  }
  else
    lvo_error(NO_MEMORY);
  
  /*
   * set asciiTextWidgets to readonly
   */
  
  XtVaSetValues(native,
		XtNeditType, XawtextRead,
		XtNdisplayCaret, False,
		NULL);
  XtVaSetValues(foreign,
		XtNeditType, XawtextRead,
		XtNdisplayCaret, False,
		NULL);
  
  XtDestroyWidget(w);            /* destroy 'apply' button */
  DisplayNodeNative(current);
  DisplayNodeForeign(current);
  SetAllSensitive((Boolean) True);
  XtOverrideTranslations(form, XtParseTranslationTable(transREV));
  XtOverrideTranslations(form, XtParseTranslationTable(transFWD));
}

/**************************************************
 *            begin of ChangePair()               *
 **************************************************/
void ChangePair(void) {
  if (current != NULL) {
    String transNREV = "<Key>Left: Noop()";
    String transNFWD = "<Key>Right: Noop()";
    Widget apply = XtVaCreateWidget("Apply changes", commandWidgetClass, form, NULL);
    XtAddCallback(apply, XtNcallback, ApplyChanges, 0);
    XtVaSetValues(apply,
		  XtNfromHoriz, test,
		  XtNhorizDistance, 50,
		  NULL);
    XtManageChild(apply);
    SetAllSensitive((Boolean) False);
    XtOverrideTranslations(form, XtParseTranslationTable(transNREV));
    XtOverrideTranslations(form, XtParseTranslationTable(transNFWD));
    
    /*
     * set asciiTextWidgets to edit mode
     */
    
    XtVaSetValues(native,
		  XtNeditType, XawtextEdit,
		  XtNdisplayCaret, True,
		  NULL);
    XtVaSetValues(foreign,
		  XtNeditType, XawtextEdit,
		  XtNdisplayCaret, True,
		  NULL);
  }
}

/************************************************************
 * FindPair : popup window to insert string and mode, then  *
 *            try to find the string from the current pair  *
 *            of words to end of wordlist                   *
 *                                                          *
 * mode : FOREIGN or NATIVE language search mode            *
 ************************************************************/
void FindPair(unsigned char mode) {
  char *search_string;
  Widget psh, form, mess, native, foreign;
  if (current != NULL) {
    psh = ComputePsh();
    
    /*
     * this is the find formWidget
     */
    
    form = XtVaCreateWidget("", formWidgetClass, psh, NULL);
    
    /*
     * this is the message labelWidget
     */
    
    mess = XtVaCreateManagedWidget("", labelWidgetClass, form, NULL);
    XtVaSetValues(mess,
		  XtNlabel, "Please insert your search string :\n\n",
		  XtNborderWidth, 0,
		  NULL);
    
    /*
     * this is the asciiTextWidget to enter the search string
     */
    
    if (mode == NATIVE) {
      native = XtVaCreateManagedWidget("native", asciiTextWidgetClass, form, NULL);
      XtVaSetValues(native,
		    XtNwidth, 200,
		    XtNfromVert, mess,
		    NULL);
      XtManageChild(form);
      XtPopup(psh, XtGrabNonexclusive);
      XtSetKeyboardFocus(psh, native);
      search_string = GetString(native);
    }
    else {
      foreign = XtVaCreateManagedWidget("foreign", asciiTextWidgetClass, form, NULL);
      XtVaSetValues(foreign,
		    XtNwidth, 200,
		    XtNfromVert, mess,
		    NULL);
      XtManageChild(form);
      XtPopup(psh, XtGrabNonexclusive);
      XtSetKeyboardFocus(psh, foreign);
      search_string = GetString(foreign);
    }
    find_pair(search_string, mode);
    free(search_string);
    XtDestroyWidget(psh);
    SetAllSensitive((Boolean) True);
  }
  else
    lvo_error(SEARCH_EMPTY_LIST);
}
