/************************************************************
 *  Martin Toepfer     ***     PROJECT : Lingvisto          *
 *  Wissmannstr. 5     ***                                  *
 *  D - 22041 Hamburg  ***     File    : test.c             *
 *  Germany            ***     Date    : 04.04.1997         *
 ************************************************************/
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "datastructure.h"
#include "edit.h"
#include "error.h"
#include "random.h"
#include "test.h"
#include "xawfunction.h"

/************************************************************
 *           global variables (in datastructure.c)          *
 ************************************************************/
extern struct lvo_file head;
extern NODE current;

/************************************************************
 *               global variables (in xaw.c)                *
 ************************************************************/
extern jmp_buf env2;

/************************************************************
 *        some more global variables and constants          *
 ************************************************************/
#define KNOWN (unsigned char) 1
#define UNKNOWN (unsigned char) 0
unsigned short known;
unsigned short unknown;
static unsigned char res;
static char *wlist[4];

/************************************************************
 *                Xaw function prototypes                   *
 ************************************************************/
static void DisplayRightNode(NODE n);
static void DisplayChoice(NODE n, unsigned char mode);

/**********************************************************************
 *                          test functions                            *
 **********************************************************************/
/************************************************************
 * restore_list : destroy cycle, rebuild linear list        *
 ************************************************************/
void restore_list(void) {
  if (get_wordlist() != NULL) {
    current = get_wordlist();
    current->prev = NULL;                  /* destroy cycle */
    
    /*
     * rebuild linear list
     */
    
    for(; current->next != NULL; current = current->next)
      current->next->prev = current;
    
    /*
     * set current to begin of wordlist
     */
    
    current = get_wordlist();
  }
}

/**********************************************************************
 *           function to compute the multiple choice test             *
 **********************************************************************/
static void choose(NODE node, unsigned char mode) {
  unsigned char r = (unsigned char)(lvo_rand() % 4);
  NODE h1 = current;
  NODE h2, h3;
  register unsigned char i;
  for(i = (unsigned char) 0; i < 4 ; i++)
    wlist[i] = NULL;

  /*
   * find first choice
   */
  
  for(i = (unsigned char) 0; i < (r + 1); i++)
    (h1->next != NULL) ? (h1 = h1->next) : (h1 = get_wordlist());
  if (h1 == node)
    (h1->next != NULL) ? (h1 = h1->next) : (h1 = get_wordlist());
  
  /*
   * find second choice
   */
  
  h2 = h1;
  for(i = (unsigned char) 0; i < 2; i++)
    (h2->next != NULL) ? (h2 = h2->next) : (h2 = get_wordlist());
  while ((h2 == h1) || (h2 == node))
    (h2->next != NULL) ? (h2 = h2->next) : (h2 = get_wordlist());
  
  /*
   * find third choice
   */

  h3 = h2;
  while ((h3 == h2) || (h3 == h1) || (h3 == node))
    (h3->next != NULL) ? (h3 = h3->next) : (h3 = get_wordlist());

  if (mode == MFOREIGN) {
    
    /*
     * ask foreign mode
     */
    
    wlist[r] = node->pair.foreign;
    for(i = (unsigned char) 0; wlist[i] != NULL; i++)
      ;
    wlist[i] = h3->pair.foreign;
    for(; wlist[i] != NULL; i++)
      ;
    wlist[i] = h1->pair.foreign;
    for(; wlist[i] != NULL; i++)
      ;
    wlist[i] = h2->pair.foreign;
  }
  else {
    
    /*
     * ask native mode
     */

    wlist[r] = node->pair.native;
    for(i = (unsigned char) 0; wlist[i] != NULL; i++)
      ;
    wlist[i] = h3->pair.native;
    for(;wlist[i] != NULL; i++)
      ;
    wlist[i] = h1->pair.native;
    for(;wlist[i] != NULL; i++)
      ;
    wlist[i] = h2->pair.native;
  }

  /*
   * now popup choose window
   */

  DisplayChoice(node, mode);
}

/**********************************************************************
 *               function to compute the regular test                 *
 **********************************************************************/
/************************************************************
 * askit : ask pair in chosen mode                          *
 *                                                          *
 * node : ask this pair of words                            *
 * mode : ask either native->FOREIGN or foreign->NATIVE     *
 ************************************************************/
static void askit(NODE node, unsigned char mode) {

  /*
   * ask foreign mode
   */
   
  if (mode == FOREIGN) {
    char *input;
    DisplayNodeNative(node);
    input = GetForeign();
    
    /*
     * known
     */
    
    if (strcmp(input, node->pair.foreign) == 0) {
      DisplayState("Right");
      
      /*
       * delete known pair from cycle
       */
      
      current->prev = current->prev->prev;
      free(input);
      res = KNOWN;
      return;
    }
    
    /*
     * unknown
     */
    
    else {
      DisplayState("Wrong");
      DisplayRightNode(node);
      free(input);
      res = UNKNOWN;
      return;
    }
  }
  
  /*
   * ask native mode
   */
  
  else {
    char *input;
    DisplayNodeForeign(node);
    input = GetNative();

    /*
     * known
     */
    
    if (strcmp(input, node->pair.native) == 0) {
      DisplayState("Right");
      
      /*
       * delete known pair from cycle
       */
      
      current->prev = current->prev->prev;
      free(input);
      res = KNOWN;
      return;
    }
    
    /*
     * unknown
     */
    
    else {
      DisplayState("Wrong");
      DisplayRightNode(node);
      free(input);
      res = UNKNOWN;
      return;
    }     
  }
}

/**********************************************************************
 *                     general test function                          *
 **********************************************************************/
/************************************************************
 * joseph : ask all pairs in chosen mode                    *
 *                                                          *
 * mode : ask either native->FOREIGN or foreign->NATIVE     *
 ************************************************************/
/************************************************************
 * IDEA : A modified 'problem of josephus' algorithm. The   *
 *        distance between two elements is a random number. *
 ************************************************************/
void joseph(unsigned char mode) {
  if (setjmp(env2) == 0) {
    NODE h;
    if ((h = get_wordlist()) != NULL) {
      unsigned char mc = (unsigned char)(mode % 2);

      /*
       * if we do a multiple choice test, the length of the
       * wordlist has to be four at least
       */
      
      if (mc != 0) {
	register unsigned char i;
	for(i = (unsigned char) 0; i < 3; i++) {
	  if (h->next != NULL)
	    h = h->next;
	  else {
	    lvo_error(SHORT_LIST);
	    return;
	  }
	}
      }

      known = 0;
      unknown = 0;
      SetAllSensitive((Boolean) False);
      
      /*
       * create cycle from linear list
       */
      
      while (current->next != NULL)
	forward_pair();
      head.wordlist->prev = current;
      
      /*
       * begin test
       */
      
      while (current->prev != current) {
	register unsigned char i;
	unsigned char r = lvo_rand();  /* get random number */
	
	/*
	 * goto next word
	 */
	
	for(i = (unsigned char) 0; i <= r; i++)
	  current = current->prev;
	
	/*
	 * ask next pair of words
	 */

	if (mc != 0) {
	  
	  /*
	   * multiple choice test
	   */

	  choose(current->prev, mode);
	  if (res == KNOWN)
	    known ++;
	  else
	    unknown++;
	}
	else {

	  /*
	   * regular test
	   */

	  askit(current->prev, mode);
	  if (res == KNOWN)
	    known++;
	  else
	    unknown++;
	}
      }
      
      /*
       * ask last word until you've known it
       */
      
      if (mc != 0) {
	
	/*
	 * multiple choice test
	 */

	do {
	  choose(current, mode);
	  if (res == KNOWN)
	    known++;
	  else
	    unknown++;
	} while (res == UNKNOWN);
      }
      else {
	
	/*
	 * regular test
	 */
	
	do {
	  askit(current, mode);
	  if (res == KNOWN)
	    known++;
	  else
	    unknown++;
	} while (res == UNKNOWN);
      }
      
      /*
       * clean up and do statistics
       */
      
      restore_list();
      DisplayResult(known, unknown);
    }
    else
      lvo_error(NO_QUESTION);
  }
}

/**********************************************************************
 *                         Xaw functions                              *
 **********************************************************************/
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/List.h>

extern Boolean b;
extern String trans4;
extern Widget top;

/************************************************************
 * itoa : simple version, converts (int) -> (char *)        *
 ************************************************************/
static char *itoa(unsigned short i, char s[]) {
  register char j = (char) 0;
  register char k, c;
  do {
    s[j++] = '0' + i % 10;
  } while ((i /= 10) > 0);
  s[j] = '\0';
  for(k = (char) 0, j = (char) strlen(s) - 1; k < j; k++, j--) {
    c = s[k];
    s[k] = s[j];
    s[j] = c;
  }
  return(s);
}

/************************************************************
 * DisplayResult : called after test to show the result     *
 ************************************************************/
void DisplayResult(unsigned short known, unsigned short unknown) {
  char *result;
  char s[6];
  char tmp[300]="Known   : ";
  
  /*
   * compute the result
   */
  
  if (unknown >= (0.7 * known))
    result = "Forget it !\n";                      /* 30 % */ 
  else if (unknown > (0.5 * known))
    result = "Unsatisfactory !\n";                 /* 50 % */
  else if (unknown > (0.3 * known))
    result = "Sufficient !\n";                     /* 70 % */
  else if (unknown > (0.15 * known))
    result = "Fair !\n";                           /* 85 % */
  else if (unknown > (0.05 * known))
    result = "Good !\n";                           /* 95 % */
  else
    result = "Very good !\n";
  
  /*
   * compute result-string
   */
  
  strcat(tmp, itoa(known, s));
  strcat(tmp, "\n\nUnknown : ");
  strcat(tmp, itoa(unknown, s));
  strcat(tmp, "\n\nResult  : ");
  strcat(tmp, result);
  DisplayMessage(tmp);
  DisplayState(get_filename());
}

/************************************************************
 * DisplayRightNode : shows correct answer in popup shell   *
 *                                                          *
 * n : node that contains the right answer                  *
 ************************************************************/
/*************************************************
 *        'ok' button callback function          *
 *************************************************/
/* ARGSUSED */
void Popdown(Widget w, XtPointer x, XtPointer y) {
  XtDestroyWidget(XtParent(XtParent(w)));
  b = (Boolean) False;
  XtSetKeyboardFocus(top, top);
}

/*************************************************
 *         begin of DisplayRightNode()           *
 *************************************************/
static void DisplayRightNode(NODE n) {
  Widget psh, form, error, native, foreign, ok;
  psh = ComputePsh();
  b = (Boolean) True;
  
  /*
   * this is the message formWidget
   */
  
  form = XtVaCreateWidget("", formWidgetClass, psh, NULL);
  
  /*
   * this is the labelWidget to display the error message
   */
  
  error = XtVaCreateManagedWidget("", labelWidgetClass, form, NULL);
  XtVaSetValues(error,
		XtNlabel, "ERROR\n\nThis is the correct answer :\n\n",
		XtNborderWidth, 0,
		NULL);
  
  /*
   * this is the labelWidget to display the native language word
   */
  
  native = XtVaCreateManagedWidget("native", labelWidgetClass, form, NULL);
  XtVaSetValues(native,
		XtNlabel, n->pair.native,
		XtNborderWidth, 0,
		XtNfromVert, error,
		NULL);
  
  /*
   * this is the labelWidget to display the foreign language word
   */
  
  foreign = XtVaCreateManagedWidget("foreign", labelWidgetClass, form, NULL);
  XtVaSetValues(foreign,
		XtNlabel, n->pair.foreign,
		XtNborderWidth, 0,
		XtNfromVert, native,
		NULL);
  
  /*
   * this is the ok commandWidget
   */
  
  ok = XtVaCreateManagedWidget("ok", commandWidgetClass, form, NULL);
  XtAddCallback(ok, XtNcallback, Popdown, NULL);
  XtVaSetValues(ok, XtNfromVert, foreign, NULL);
  XtOverrideTranslations(ok, XtParseTranslationTable(trans4));
  
  /*
   * now lets popup the shell
   */
  
  XtManageChild(form);
  XtPopup(psh, XtGrabExclusive);
  XtSetKeyboardFocus(psh, ok);
  XtSetKeyboardFocus(top, ok);
  
  /*
   * wait until 'ok' is pressed
   */
  
  while (b == (Boolean) True)
    DoNextEvent();
}

/************************************************************
 * DisplayChoice : popup shell for multiple choice test     *
 *                                                          *
 * n : node that contains the right answer                  *
 ************************************************************/
/*************************************************
 *            listWidget callback                *
 *************************************************/
/* ARGSUSED */
static void SelectN(Widget w, XtPointer x, XtPointer y) {
  NODE n = (NODE) x;
  XawListReturnStruct *s = (XawListReturnStruct *) y;
  if (strcmp(n->pair.native, s->string) != 0) {
    res = UNKNOWN;
    DisplayRightNode(n);
  }
  else {
    current->prev = current->prev->prev;
    res = KNOWN;
  }
  b = (Boolean) False;
  Destroy(w, x, y);
}

/*************************************************
 *            listWidget callback                *
 *************************************************/
/* ARGSUSED */
static void SelectF(Widget w, XtPointer x, XtPointer y) {
  NODE n = (NODE) x;
  XawListReturnStruct *s = (XawListReturnStruct *) y;
  if (strcmp(n->pair.foreign, s->string) != 0) {
    res = UNKNOWN;
    DisplayRightNode(n);
  }
  else {
    current->prev = current->prev->prev;
    res = KNOWN;
  }
  b = (Boolean) False;
  Destroy(w, x, y);
}

/*************************************************
 *          commandWidget callback               *
 *************************************************/
void JosephExit(Widget w, XEvent *e, String *p, Cardinal *n);
/* ARGSUSED */
static void QuitJ(Widget w, XtPointer x, XtPointer y) {
  Destroy(w, NULL, NULL);
  JosephExit(w, NULL, NULL, NULL);
}

/*************************************************
 *         begin of DisplayChoice()              *
 *************************************************/
static void DisplayChoice(NODE n, unsigned char mode) {
  Widget foreign, form, l, native, psh, quit;
  psh = ComputePsh();
  b = (Boolean) True;

  /*
   * this is the multiple choice formWidget
   */
  
  form = XtVaCreateWidget("form", formWidgetClass, psh, NULL);

  /*
   * this is the 'Multiple-Choice-Test' labelWidget
   */

  l = XtVaCreateManagedWidget("Multiple-Choice-Test", labelWidgetClass, form, NULL);

  /*
   * ask foreign mode
   */

  if (mode == MFOREIGN) {

    /*
     * this is the native question labelWidget
     */

    native = XtVaCreateManagedWidget("native", labelWidgetClass, form, NULL);
    XtVaSetValues(native,
		  XtNfromVert, l,
		  XtNvertDistance, 20,
		  XtNlabel, n->pair.native,
		  NULL);

    /*
     * this is the foreign answer listWidget
     */

    foreign = XtVaCreateManagedWidget("foreign", listWidgetClass, form, NULL);
    XawListChange(foreign, wlist, 4, -1, True);
    XtAddCallback(foreign, XtNcallback, SelectF, (XtPointer) n);
    XtVaSetValues(foreign, 
		  XtNdefaultColumns, 1,
		  XtNrowSpacing, 4,
		  XtNfromVert, native,
		  NULL);

    /*
     * this is the quit commandWidget
     */

    quit = XtVaCreateManagedWidget("", commandWidgetClass, form, NULL);
    XtVaSetValues(quit,
		  XtNlabel, "Quit Test",
		  XtNfromVert, foreign,
		  XtNvertDistance, 20,
		  NULL);
  }
  else {

    /*
     * this is the foreign question labelWidget
     */

    foreign = XtVaCreateManagedWidget("foreign", labelWidgetClass, form, NULL);
    XtVaSetValues(foreign,
		  XtNfromVert, l,
		  XtNvertDistance, 20,
		  XtNlabel, n->pair.foreign,
		  NULL);

    /*
     * this is the native answer listWidget
     */

    native = XtVaCreateManagedWidget("native", listWidgetClass, form, NULL);
    XawListChange(native, wlist, 4, -1, True);
    XtAddCallback(native, XtNcallback, SelectN, (XtPointer) n);
    XtVaSetValues(native,
		  XtNdefaultColumns, 1,
		  XtNrowSpacing, 4,
		  XtNfromVert, foreign,
		  NULL);

    /*
     * this is the quit commandWidget
     */

    quit = XtVaCreateManagedWidget("", commandWidgetClass, form, NULL);
    XtVaSetValues(quit,
		  XtNlabel, "Quit Test",
		  XtNfromVert, native,
		  XtNvertDistance, 20,
		  NULL);
  }
  XtAddCallback(quit, XtNcallback, QuitJ, NULL);
  XtManageChild(form);
  XtPopup(psh, XtGrabNonexclusive);
  while (b == (Boolean) True)
    DoNextEvent();
}

